#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <malloc.h>
#include "dynamic_libs/os_functions.h"
#include "dynamic_libs/fs_functions.h"
#include "dynamic_libs/gx2_functions.h"
#include "dynamic_libs/sys_functions.h"
#include "dynamic_libs/vpad_functions.h"
#include "dynamic_libs/padscore_functions.h"
#include "dynamic_libs/socket_functions.h"
#include "dynamic_libs/ax_functions.h"
#include "fs/fs_utils.h"
#include "fs/sd_fat_devoptab.h"
#include "system/memory.h"
#include "utils/logger.h"
#include "utils/utils.h"
#include "common/common.h"

int screen_buf0_size = 0;
int screen_buf1_size = 0;

//Convenience!
void _printstr(int line, char* string) {
	// print to TV
    OSScreenPutFontEx(0, 0, line, string);

    // print to DRC
    OSScreenPutFontEx(1, 0, line, string);

    // Flush the cache
    DCFlushRange((void*)0xF4000000, screen_buf0_size);
    DCFlushRange(((void*)0xF4000000 + screen_buf0_size), screen_buf1_size);

    // Flip buffers
    OSScreenFlipBuffersEx(0);
    OSScreenFlipBuffersEx(1);
	
	// print to TV
    OSScreenPutFontEx(0, 0, line, string);

    // print to DRC
    OSScreenPutFontEx(1, 0, line, string);

    // Flush the cache
    DCFlushRange((void*)0xF4000000, screen_buf0_size);
    DCFlushRange(((void*)0xF4000000 + screen_buf0_size), screen_buf1_size);

    // Flip buffers
    OSScreenFlipBuffersEx(0);
    OSScreenFlipBuffersEx(1);
}

/* Entry point */
int Menu_Main(void)
{
	//This is all standard stuff, skip down to the #defines
	
    //!*******************************************************************
    //!                   Initialize function pointers                   *
    //!*******************************************************************
    //! do OS (for acquire) and sockets first so we got logging
    InitOSFunctionPointers();
    InitSocketFunctionPointers();

    //log_init("192.168.178.3");
    //log_print("Starting launcher\n");

    InitFSFunctionPointers();
    InitVPadFunctionPointers();

    //log_print("Function exports loaded\n");

    //!*******************************************************************
    //!                    Initialize heap memory                        *
    //!*******************************************************************
    //log_print("Initialize memory management\n");
    //! We don't need bucket and MEM1 memory so no need to initialize
    memoryInitialize();

    //!*******************************************************************
    //!                        Initialize FS                             *
    //!*******************************************************************
    //log_printf("Mount SD partition\n");
    mount_sd_fat("sd");

    VPADInit();
	
    // Init screen and screen buffers
    OSScreenInit();
    screen_buf0_size = OSScreenGetBufferSizeEx(0);
    screen_buf1_size = OSScreenGetBufferSizeEx(1);

    OSScreenSetBufferEx(0, (void*)0xF4000000);
    OSScreenSetBufferEx(1, ((void*)0xF4000000 + screen_buf0_size));

    OSScreenEnableEx(0, 1);
    OSScreenEnableEx(1, 1);

    // Clear screens
    OSScreenClearBufferEx(0, 0);
    OSScreenClearBufferEx(1, 0);
	
	_printstr(0, "Say HELLO to the IOSU! v0.69");
    
	#define RET_SIZE 0x70 //Size of the return value from our IOCTL. More on that later.
	#define IOCTL_NUM 0x4 //Command number to send to node. Check wiiubrew.
	#define A_OFFSET 0x24 //Info of the stuff we want from the big mess the IOSU will throw at us.
	#define A_SIZE 0x20
	#define B_OFFSET 0x4E
	#define B_SIZE 0x40 //This is too big.
	u32 coreinit_handle = 0;
	unsigned int (*IOS_Open)(const char* path, int mode);
	unsigned int (*IOS_Ioctl)(int fd, int cmd, void* inbuf, size_t inlen, void* outbuf, size_t outlen);
	unsigned int (*IOS_Close)(int fd);
	OSDynLoad_Acquire("coreinit", &coreinit_handle);
	OSDynLoad_FindExport(coreinit_handle, 0, "IOS_Open", &IOS_Open);
	OSDynLoad_FindExport(coreinit_handle, 0, "IOS_Ioctl", &IOS_Ioctl);
	OSDynLoad_FindExport(coreinit_handle, 0, "IOS_Close", &IOS_Close);
	
	//Get a handle to the path we want. This one has lax permissions and a comical payload.
	unsigned int fd = IOS_Open("/dev/net/ifmgr/ncl", 1);
	
	char buf[255]; //snprintf buffer
	
	//Tell the user our handle, because why not.
	__os_snprintf(buf, 255, "Hello IOSU! fd %08X. Leaking...", fd);
	_printstr(1, buf);
	
	//Allocate a buffer for the IOSU to shove data into.
	unsigned char* rawConfig = malloc(RET_SIZE);
	memset(rawConfig, 0, RET_SIZE);
	
	//Send a message to the IOSU asking it to run IOCTL_NUM.
	int ret = IOS_Ioctl(fd, IOCTL_NUM, 0, 0, rawConfig, RET_SIZE);
	
	//Tell the user
	__os_snprintf(buf, 255, "Leak complete! Return code %X.", ret);
	_printstr(2, buf);
	
	//Simple stringbuilder. The strings I want are meant to be null terminated but snprintf didn't like it.
	char a[A_SIZE];
	__os_snprintf(a, A_SIZE, ""); //Empty out a
	for (int i = 0; i < A_SIZE; i++) { //For each character in big A...
		if (*(rawConfig + A_OFFSET + i) == 0) { //If it's a null character
			break; //Stop
		} //Otherwise...
		__os_snprintf(a, A_SIZE, "%s%c", a, *(rawConfig + A_OFFSET + i)); //Add it to little a
	}
	
	char b[B_SIZE];
	__os_snprintf(b, B_SIZE, "");
	for (int i = 0; i < B_SIZE; i++) {
		if (*(rawConfig + B_OFFSET + i) == 0) {
			break;
		}
		__os_snprintf(b, B_SIZE, "%s%c", b, *(rawConfig + B_OFFSET + i));
	}
	
	//Take the user for a bit of a ride
	_printstr(4, "IOSU communication complete! Press A to see what we found.");
	_printstr(5, "The data could be anything. You have been warned.");
	
	int vpadError = -1;
    VPADData vpad;
	
	while(1)
    {
        VPADRead(0, &vpad, 1, &vpadError);

        if(vpadError == 0 && ((vpad.btns_d | vpad.btns_h) & VPAD_BUTTON_A))
            break;

		usleep(500);
    }
	
	//print our data
	
	_printstr(7, a);
	_printstr(8, b);
	
	_printstr(10, "WHO IS YOUR SAVIOUR NOW"); //run the program, I dare you
	
	//Tidy up - wipe and free the IOSU buffer
	memset(rawConfig, 0, RET_SIZE);
	free(rawConfig);
	
	//Close our IOSU handle
	ret = IOS_Close(fd);
	
	//Tell the user
	__os_snprintf(buf, 255, "Closed IOSU handle. Code %X.", ret);
	
	_printstr(16, buf);
	_printstr(17, "Press HOME to quit.");
	
    while(1)
    {
        VPADRead(0, &vpad, 1, &vpadError);

        if(vpadError == 0 && ((vpad.btns_d | vpad.btns_h) & VPAD_BUTTON_HOME))
            break;

		usleep(500);
    }

    //!*******************************************************************
    //!                    Enter main application                        *
    //!*******************************************************************

    //log_printf("Unmount SD\n");
    unmount_sd_fat("sd");
    //log_printf("Release memory\n");
    memoryRelease();
    //log_deinit();

    return EXIT_SUCCESS;
}

